use sqlx::{PgPool, Postgres};

use crate::model::member::Roles;

#[derive(validator::Validate)]
pub struct User {
    pub user_id: uuid::Uuid,
    #[validate(email)]
    pub email: String,
    pub password: String,
    pub admin: bool,
}

impl User {
    pub async fn insert(
        transaction: &mut sqlx::Transaction<'_, Postgres>,
        email: &str,
        password: &str,
    ) -> Result<uuid::Uuid, sqlx::Error> {
        let user_id = uuid::Uuid::new_v4();

        sqlx::query!(
            "
                INSERT INTO users (
                    user_id, email, password, admin
                )
                VALUES (
                    $1, $2, $3, $4
                );
            ",
            &user_id,
            email,
            password,
            false
        )
        .execute(&mut **transaction)
        .await?;

        Ok(user_id)
    }

    pub async fn get_from_email(transaction: &PgPool, email: String) -> Result<Self, sqlx::Error> {
        let user = sqlx::query_as!(Self, "SELECT * FROM users WHERE email = $1", email)
            .fetch_one(transaction)
            .await?;

        Ok(user)
    }

    pub async fn get(transaction: &PgPool, user_id: uuid::Uuid) -> Result<Self, sqlx::Error> {
        let user = sqlx::query_as!(Self, "SELECT * FROM users WHERE user_id = $1", user_id)
            .fetch_one(transaction)
            .await?;

        Ok(user)
    }
}

#[derive(Debug)]
pub struct UserMember {
    user_id: uuid::Uuid,
    member_id: String,
}

impl UserMember {
    pub async fn insert_many(
        transaction: &mut sqlx::Transaction<'_, Postgres>,
        user_ids: &Vec<uuid::Uuid>,
        member_ids: &Vec<String>,
    ) -> Result<(), sqlx::Error> {
        sqlx::query!(
            "
                INSERT INTO users_members (
                    user_id, member_id
                )
                SELECT * FROM UNNEST($1::uuid[], $2::varchar[])
            ",
            &user_ids[..],
            &member_ids[..]
        )
        .execute(&mut **transaction)
        .await?;

        Ok(())
    }

    pub async fn get_roles(pool: &PgPool, user_id: &uuid::Uuid) -> Result<Roles, sqlx::Error> {
        let roles = sqlx::query_scalar!(
            "
                SELECT roles FROM users_members INNER JOIN members ON users_members.member_id = members.member_id AND users_members.user_id = $1;
            ",
            user_id
        ).fetch_all(pool).await?;

        let roles: Vec<Roles> = roles.into_iter().map(|r| r.into()).collect();
        let roles = roles
            .into_iter()
            .fold(Roles::empty(), |acc, flag| acc | flag);

        Ok(roles)
    }
}
